package com.example.blt_nfc

import android.app.PendingIntent
import android.content.BroadcastReceiver
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.content.pm.PackageManager
import android.nfc.NdefMessage
import android.nfc.NdefRecord
import android.nfc.NfcAdapter
import android.nfc.Tag
import android.nfc.tech.Ndef
import android.os.Build
import android.os.Bundle
import android.widget.Toast
import androidx.activity.ComponentActivity
import androidx.activity.compose.setContent
import androidx.annotation.RequiresApi
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.width
import androidx.compose.foundation.rememberScrollState
import androidx.compose.foundation.text.selection.SelectionContainer
import androidx.compose.foundation.verticalScroll
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Check
import androidx.compose.material.icons.filled.Close
import androidx.compose.material3.AssistChip
import androidx.compose.material3.AssistChipDefaults
import androidx.compose.material3.Button
import androidx.compose.material3.Divider
import androidx.compose.material3.Icon
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.OutlinedTextField
import androidx.compose.material3.Surface
import androidx.compose.material3.Text
import androidx.compose.runtime.collectAsState
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.text.font.FontWeight
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp
import androidx.lifecycle.viewmodel.compose.viewModel
import com.example.blt_nfc.ui.theme.BLT_NFCTheme
import java.nio.charset.Charset


fun ByteArray.toHexString(): String {
    val hexChars = "0123456789ABCDEF"
    val result = StringBuilder(size * 2)

    map { byte ->
        val value = byte.toInt()
        val hexChar1 = hexChars[value shr 4 and 0x0F]
        val hexChar2 = hexChars[value and 0x0F]
        result.append(hexChar1)
        result.append(hexChar2)
    }

    return result.toString()
}

@Suppress("DEPRECATION")
class MainActivity : ComponentActivity() {
    private var nfcAdapter: NfcAdapter? = null
    private lateinit var appViewModel: NFCViewModel

    @RequiresApi(Build.VERSION_CODES.JELLY_BEAN_MR2)
    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        val isNFCAvailable = packageManager.hasSystemFeature(PackageManager.FEATURE_NFC)
        // Get the default NFC adapter
        if (isNFCAvailable) nfcAdapter = NfcAdapter.getDefaultAdapter(this)
        createNFCIntentFilter()

        val filter = IntentFilter(NfcAdapter.ACTION_ADAPTER_STATE_CHANGED)
        this.registerReceiver(mReceiver, filter)


        setContent {
            appViewModel = viewModel()
            val uiState = appViewModel.state.collectAsState()

            appViewModel.updateIsNFCAvailable(isNFCAvailable)
            if (isNFCAvailable) appViewModel.updateIsNFCEnabled(nfcAdapter?.isEnabled == true)

            BLT_NFCTheme {
                // A surface container using the 'background' color from the theme
                Surface(
                    modifier = Modifier.fillMaxSize(), color = MaterialTheme.colorScheme.background
                ) {
                    Column(modifier = Modifier.padding(horizontal = 16.dp, vertical = 8.dp)) {
                        if (!uiState.value.isNFCAvailable) Text(text = "Sorry! Your phone does not support NFC.")
                        if (uiState.value.isNFCAvailable) {
                            Row(
                                modifier = Modifier.fillMaxWidth(1f),
                                verticalAlignment = Alignment.CenterVertically,
                                horizontalArrangement = Arrangement.SpaceBetween
                            ) {
                                Text(
                                    text = "NFC status",
                                    fontSize = 18.sp,
                                    fontWeight = FontWeight.Bold
                                )
                                Row(horizontalArrangement = Arrangement.spacedBy(8.dp)) {
                                    AssistChip(
                                        onClick = { /*TODO*/ },
                                        label = { Text(text = "Available") },
                                        leadingIcon = {
                                            Icon(
                                                imageVector = if (uiState.value.isNFCAvailable) Icons.Filled.Check else Icons.Filled.Close,
                                                contentDescription = if (uiState.value.isNFCAvailable) "NFC is available" else "NFC is not available"
                                            )
                                        },
                                        colors = AssistChipDefaults.assistChipColors(containerColor = if (uiState.value.isNFCAvailable) MaterialTheme.colorScheme.background else MaterialTheme.colorScheme.errorContainer)
                                    )
                                    if (uiState.value.isNFCAvailable) {
                                        AssistChip(
                                            onClick = { /*TODO*/ },
                                            label = { Text(text = if (uiState.value.isNFCEnabled) "Enabled" else "Disabled") },
                                            leadingIcon = {
                                                Icon(
                                                    imageVector = if (uiState.value.isNFCEnabled) Icons.Filled.Check else Icons.Filled.Close,
                                                    contentDescription = if (uiState.value.isNFCEnabled) "NFC is enabled" else "NFC is not enabled",
                                                    tint = if (uiState.value.isNFCEnabled) MaterialTheme.colorScheme.primary else MaterialTheme.colorScheme.onErrorContainer
                                                )
                                            },
                                            colors = AssistChipDefaults.assistChipColors(
                                                containerColor = if (uiState.value.isNFCEnabled) MaterialTheme.colorScheme.background else MaterialTheme.colorScheme.errorContainer
                                            )
                                        )
                                    }
                                }

                            }

                            Row(
                                modifier = Modifier.fillMaxWidth(1f),
                                verticalAlignment = Alignment.CenterVertically,
                                horizontalArrangement = Arrangement.SpaceBetween
                            ) {
                                Text(
                                    text = "Mode: ${if (uiState.value.isWriting) "Write" else "Read"}",
                                    fontSize = 18.sp,
                                    fontWeight = FontWeight.Bold
                                )
                                Row {
                                    Button(onClick = { appViewModel.updateIsWriting(!uiState.value.isWriting) }) {
                                        Text(text = if (uiState.value.isWriting) "Read" else "Write")
                                    }
                                    Spacer(modifier = Modifier.width(8.dp))
                                    Button(onClick = { appViewModel.clearState() }) {
                                        Text(text = "Reset")
                                    }
                                }
                            }

                            if (uiState.value.isWriting) {
                                OutlinedTextField(
                                    value = uiState.value.messageToWrite,
                                    onValueChange = { appViewModel.updateMessageToWrite(it) },
                                    label = { Text("Payload to write") },
                                    maxLines = 3,
                                    modifier = Modifier.fillMaxWidth(1f)
                                )
                            }
                            Divider(modifier = Modifier.padding(top = 16.dp, bottom = 4.dp))

                            SelectionContainer {
                                Column(modifier = Modifier.verticalScroll(state = rememberScrollState())) {

                                    // TechList
                                    Text(text = getSection("TechList", uiState.value.techList))
                                    Divider(modifier = Modifier.padding(vertical = 4.dp))
                                    // Records
                                    Text(text = getSection("Records", uiState.value.records))
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    @RequiresApi(Build.VERSION_CODES.JELLY_BEAN)
    override fun onNewIntent(intent: Intent?) {
        super.onNewIntent(intent)

        val tag: Tag? = intent?.getParcelableExtra(NfcAdapter.EXTRA_TAG)
        if (!appViewModel.state.value.isWriting) { // read from tag
            val techListString = tag?.techList?.joinToString() ?: ""
            appViewModel.updateTechList(techListString)

            if (NfcAdapter.ACTION_TAG_DISCOVERED == intent?.action) {
                intent.getParcelableArrayExtra(NfcAdapter.EXTRA_NDEF_MESSAGES)
                    ?.also { rawMessages ->
                        val messages: List<NdefMessage> = rawMessages.map { it as NdefMessage }
                        // Process the messages array.
                        val parsedMessaged = mutableListOf<String>()
                        messages.map { msg ->
                            msg.records.map { ndefRecord ->
                                val type = ndefRecord.toMimeType()
                                val payload = ndefRecord.payload.decodeToString()
                                parsedMessaged.add("[$type]: $payload")
                            }
                        }
                        appViewModel.updateRecords(parsedMessaged.joinToString(",\n"))
                    }
            }
        }

        if (appViewModel.state.value.isWriting) { // Write to tag
            val ndef = Ndef.get(tag)
            val text = appViewModel.state.value.messageToWrite
            val record =
                NdefRecord.createMime("text/plain", text.toByteArray(Charset.forName("US-ASCII")))
            val message = NdefMessage(arrayOf(record))

            ndef?.let {
                if (it.maxSize < message.toByteArray().size) {
                    Toast.makeText(
                        this,
                        "Write failed! NFC tag does not have enough storage.",
                        Toast.LENGTH_LONG
                    ).show()
                    return@let
                }
                try {
                    it.connect()
                    it.writeNdefMessage(message)
                    it.close()
                    Toast.makeText(this, "Data Written successfully.", Toast.LENGTH_LONG).show()
                } catch (e: Exception) {
                    Toast.makeText(
                        this, "Unknown error occurred while writing data to tag.", Toast.LENGTH_LONG
                    ).show()
                }
            }
        }
    }


    override fun onResume() {
        super.onResume()
        val pendingIntent = PendingIntent.getActivity(
            this,
            0,
            Intent(this, javaClass).addFlags(Intent.FLAG_ACTIVITY_SINGLE_TOP),
            PendingIntent.FLAG_MUTABLE
        )
        val intentFilters = arrayOf(
            IntentFilter(NfcAdapter.ACTION_NDEF_DISCOVERED),
            IntentFilter(NfcAdapter.ACTION_TAG_DISCOVERED),
            IntentFilter(NfcAdapter.ACTION_TECH_DISCOVERED)
        )
        nfcAdapter?.enableForegroundDispatch(this, pendingIntent, intentFilters, null)
    }

    override fun onPause() {
        super.onPause()
        nfcAdapter?.disableForegroundDispatch(this)
    }

    override fun onDestroy() {
        super.onDestroy()
        this.unregisterReceiver(mReceiver)
    }

    private fun createNFCIntentFilter(): Array<IntentFilter> {
        val intentFilter = IntentFilter(NfcAdapter.ACTION_NDEF_DISCOVERED)
        try {
            intentFilter.addDataType("*/*")
        } catch (e: IntentFilter.MalformedMimeTypeException) {
            throw RuntimeException("Failed to add MIME type.", e)
        }
        return arrayOf(intentFilter)
    }

    private fun getLineSeparator(char: Char = '-', width: Int = 73): String {
        return char.toString().repeat(width)
    }

    private fun getSection(sectionTitle: String, sectionContent: String): String {
        val separator = getLineSeparator()
        return "$sectionTitle:\n$separator\n$sectionContent\n$separator\n"
    }

    private val mReceiver: BroadcastReceiver = object : BroadcastReceiver() {
        @RequiresApi(Build.VERSION_CODES.JELLY_BEAN_MR2)
        override fun onReceive(context: Context, intent: Intent) {
            val action = intent.action
            if (action == NfcAdapter.ACTION_ADAPTER_STATE_CHANGED) {
                val state = intent.getIntExtra(
                    NfcAdapter.EXTRA_ADAPTER_STATE, NfcAdapter.STATE_OFF
                )
                when (state) {
                    NfcAdapter.STATE_OFF -> {
                        appViewModel.updateIsNFCEnabled(false)
                    }

                    NfcAdapter.STATE_TURNING_OFF -> {}
                    NfcAdapter.STATE_ON -> {
                        appViewModel.updateIsNFCEnabled(true)
                    }

                    NfcAdapter.STATE_TURNING_ON -> {}
                }
            }
        }
    }
}
